==> fr.ac6.linux.utils_1.0.0.201706270959.jar: resources/scripts/version_utils.sh <==
#
# Compare two version strings, returning:
#		1 if V1 is greater than V2
# 		2 if V2 is greater than V1
# 		0 if V1 and V2 are equal
#
compare_versions() {
	local VR1="$1"
	local VR2="$2"
	local V1="${V1%%-*}"
	local V2="${V1%%-*}"
	__compare_versions "$V1" "$V2"
	local result=$?
	if [ $result -ne 0 ]; then return $result; fi
	local R1="${VR1#V1}"
	local R2="${VR2#V2}"
	R1="${R1#-}"
	R2="${R2#-}"
	if [ -n "$R1" ] || [ -n "$R2" ]; then
		__compare_versions "$R1" "$R2"
	else
		return 0
	fi
}

is_number() {
	case "$BASH_VERSION" in
		3.1.*)
			PATTERN='\^\[0-9\]+\$'
			;;
		*)
			PATTERN='^[0-9]+$'
			;;
	esac
	
	[[ "$1" =~ $PATTERN ]]
}

__compare_versions() {
	local V1="$1"
	local V2="$2"
	local A="${V1%%-*}"
	local B="${V2%%-*}"

	if [ "$A" == "$B" ]; then
		if [ "$A" = "$V1" ]; then
			if [ "$B" = "$V2" ]; then
				return 0
			else
				return 2
			fi
		elif [ "$B" = "$V2" ]; then
			return 1
		fi
		__compare_versions "${V1#*:}" "${V2#*:}"
	else
		if [ -z "$A" ]; then
    		if [ -z "$B" ]; then
    			return 0
    		else
    			return 2
    		fi
    	elif [ -z "$B" ]; then
    		return 1
    	fi
		if is_number "$A" && is_number "$B"; then
			if [[ "$A" -gt "$B" ]]; then
				return 1
			else
				return 2
			fi
        else
			if [[ "$A" < "$B" ]]; then
				return 2
			else
				return 1
			fi
		fi
	fi
}

version_expr() {
	local A="$1"
	local op="$2"
	local B="$3"
	local res
	compare_versions $A $B
	case $? in
		0)	res='=';;
		1)	res='>';;
		2)	res='<';;
	esac
	if [[ "$op" =~ "!" ]]; then
		[ "$res" != "=" ] && return 0 || return 1
	else
    	[[ "$op" =~ "$res" ]] && return 0 || return 1
    fi
}

provide_version() {
	local required="$1"
	shift
	local pkg_spec="$1"
	shift
	local version
	if [ -z "$*" ]; then
		if $required; then
			echo "Can't find required package matching '$pkg_spec'" >&2
			return 1
		fi
		log "Skipping optional package '$pkg_spec'" >&2
		return 0
	fi
	# Display the newest version available
	for version; do
		echo $version
	done | sort -rV | head -1
}

#
# Parse a dependency list and provide a list of existing packages to load
#
# The input is a set of package dependencies, with an optional version tag (+ optional arch) or constraint:
#
#		package-name_package-version_package-arch
#		package-name_package-version
#		package-name
#		package-name (version-constraints)
#
# Where version-constraint can be one or more constraints:
#		< version
#		<= version
#		== version (can be abridged as '= version' or just 'version')
#		!= version
#		>= version
#		> version
#
# When several constraints are given, they may be separated by
#		||
#		&& (optional)
# Note that && has higher precedence than || and parentheses are not supported
#
parse_depend() {
	local required="$1"
	shift
	local dependencies="$*"
	dependencies="${dependencies//(/ ( }"
	dependencies="${dependencies//)/ ) }"
	dependencies="${dependencies//</ < }"
	dependencies="${dependencies//>/ > }"
	dependencies="${dependencies//=/ = }"
	dependencies="${dependencies//!/ ! }"
	dependencies="${dependencies//&/ & }"
	dependencies="${dependencies//|/ | }"
	dependencies="${dependencies//< =/<=}"
	dependencies="${dependencies//> =/>=}"
	dependencies="${dependencies//= =/=}"
	dependencies="${dependencies//! =/!=}"
	dependencies="${dependencies//& &/&&}"
	dependencies="${dependencies//| |/||}"
	declare -a tokens=($dependencies)
	pkg_dbg "dependencies='$dependencies'"
	local state=init
	local token next_token i j pkg pkg_spec='' 
	local versions accumulated_versions filtered version
	local i=0
	for ((i=0; i<=${#tokens[*]}; i++)); do
		if [[ $i -lt ${#tokens[*]} ]]; then
			token="${tokens[i]}"
			j=$((i+1))
			if [[ $j -lt ${#tokens[*]} ]]; then
				next_token="${tokens[j]}"
			else
				next_token="EOF" # End of dependencies sentinel
			fi
		else
			token="EOF"
			next_token="EOF"
		fi
		pkg_dbg "versions='$versions'"
		pkg_dbg "accumulated_versions='$accumulated_versions'"
		pkg_dbg "filtered='$filtered'"
		pkg_dbg "token='$token'"
		pkg_dbg "next_token='$next_token'"
		pkg_dbg "state='$state'"
		pkg_dbg "pkg_spec='$pkg_spec'"
		case $state in
			init | parsed)
				if [ $state = parsed ]; then
					if provide_version $required $pkg_spec $versions $filtered; then
						true
					else
						return 1
					fi
					if [ "$token" = "EOF" ]; then
						break
					fi
				fi

				pkg="$token"
				pkg_spec="$pkg"
				versions=
				for version in $(search_all_versions $pkg); do
					versions="$versions $version"
				done
				versions="$versions "
				case "$token" in
					[a-zA-Z0-9]*)
						case "$next_token" in
							[a-zA-Z0-9]*)
								state=parsed
								;;
							*)
								state=constrained
								;;
						esac
						;;
					*)	echo "Syntax error in '$dependencies' at '$pkg': not a package name" >&2
						return 1
						;;
				esac
				;;
			constrained)
				op="$token"
				case "$op" in
					\()	
						pkg_spec="$pkg_spec $op"
						((level++))
						# Here to manage parentheses I should manage to_filter, filtered
						# and accumulated_versions in a stack
						to_filter="$versions"
						accumulated_versions=
						filtered=
						state=operator
						;;
					*)
						echo "Syntax error in '$pkg_spec' dependency: invalid token '$op'" >&2
						return 1
						;;
				esac
				;;
			operator)
				cmp="$token"
				case "$cmp" in
					\))	
						pkg_spec="$pkg_spec $op"
						if [ $level -eq 0 ]; then
							echo "Syntax error in '$pkg_spec' dependency: too much closing parentheses" >&2
							return 1
						fi
						# Here to manage parentheses I should manage to_filter, filtered
						# and accumulated_versions in a stack
						((level--))
						if [ $level -eq 0 ]; then
							versions="$accumulated_versions $filtered"
							state=parsed
						fi
						;;
					'<' | '<=' | '==' | '=' | '!=' | '!' | '>=' | '>')
						state=operand
						;;
					'||')
						accumulated_versions="$accumulated_versions $filtered"
						to_filter="$versions"
						filtered=
						;;
					'&&')
						;;
					*)
						cmp='='
						state=operand
						((i--))
#						echo "Syntax error in '$pkg_spec' dependency: operator needed before '$cmp'" >&2
#						return 1
						;;
				esac
				;;
			operand)
				val="$token"
				filtered=
				for version in $to_filter; do
					vers=${version#*_}
					if version_expr "${vers%%_*}" $cmp "$val"; then
						filtered="$filtered $version"
					fi
				done
				to_filter="$filtered"
				;;
		esac
	done
}

==> fr.ac6.linux.utils_1.0.0.202202211611.jar: resources/scripts/version_utils.sh <==
#
# Compare two version strings, returning:
#		1 if V1 is greater than V2
# 		2 if V2 is greater than V1
# 		0 if V1 and V2 are equal
#
compare_versions() {
	local VR1="$1"
	local VR2="$2"
	local V1="${VR1%%-*}"
	local V2="${VR1%%-*}"
	__compare_versions "$V1" "$V2"
	local result=$?
	if [ $result -ne 0 ]; then return $result; fi
	local R1="${VR1#V1}"
	local R2="${VR2#V2}"
	R1="${R1#-}"
	R2="${R2#-}"
	if [ -n "$R1" ] || [ -n "$R2" ]; then
		__compare_versions "$R1" "$R2"
	else
		return 0
	fi
}

is_number() {
	case "$BASH_VERSION" in
		3.1.*)
			PATTERN='\^\[0-9\]+\$'
			;;
		*)
			PATTERN='^[0-9]+$'
			;;
	esac
	
	[[ "$1" =~ $PATTERN ]]
}

__compare_versions() {
	local V1="$1"
	local V2="$2"
	local A="${V1%%-*}"
	local B="${V2%%-*}"

	if [ "$A" == "$B" ]; then
		if [ "$A" = "$V1" ]; then
			if [ "$B" = "$V2" ]; then
				return 0
			else
				return 2
			fi
		elif [ "$B" = "$V2" ]; then
			return 1
		fi
		__compare_versions "${V1#*-}" "${V2#*-}"
	else
		if [ -z "$A" ]; then
    		if [ -z "$B" ]; then
    			return 0
    		else
    			return 2
    		fi
    	elif [ -z "$B" ]; then
    		return 1
    	fi
		if is_number "$A" && is_number "$B"; then
			if [[ "$A" -gt "$B" ]]; then
				return 1
			else
				return 2
			fi
        else
			if [[ "$A" < "$B" ]]; then
				return 2
			else
				return 1
			fi
		fi
	fi
}

version_expr() {
	local A="$1"
	local op="$2"
	local B="$3"
	local res
	compare_versions $A $B
	case $? in
		0)	res='=';;
		1)	res='>';;
		2)	res='<';;
	esac
	if [[ "$op" =~ "!" ]]; then
		[ "$res" != "=" ] && return 0 || return 1
	else
    	[[ "$op" =~ "$res" ]] && return 0 || return 1
    fi
}

provide_version() {
	local required="$1"
	shift
	local pkg_spec="$1"
	shift
	local version
	if [ -z "$*" ]; then
		if $required; then
			echo "Can't find required package matching '$pkg_spec'" >&2
			return 1
		fi
		log "Skipping optional package '$pkg_spec'" >&2
		return 0
	fi
	# Display the newest version available
	for version; do
		echo $version
	done | sort -rV | head -1
}

get_comparison() {
	case $1 in
		# String comparisons
		*\<~*)		echo "[[ \${platform_options[${1%%<*}]} '<' \"${1#*~}\" ]]"		;;
		*\>~*)		echo "[[ \${platform_options[${1%%>*}]} '>' \"${1#*~}\" ]]"		;;
		*=~*)		echo "[[ \${platform_options[${1%%=*}]} '=' \"${1#*~}\" ]]"		;;
		
		# Version comparisons
		*\<=*.*)	echo "version_expr \${platform_options[${1%%<*}]} '<=' ${1#*=}"	;;
		*\>=*.*)	echo "version_expr \${platform_options[${1%%>*}]} '>=' ${1#*=}"	;;
		*==*.*)		echo "version_expr \${platform_options[${1%%=*}]} '==' ${1#*=}"	;;
		*\<*.*)		echo "version_expr \${platform_options[${1%%<*}]} '<'  ${1#*<}"	;;
		*\>*.*)		echo "version_expr \${platform_options[${1%%>*}]} '>'  ${1#*>}"	;;
		*=*.*)		echo "version_expr \${platform_options[${1%%=*}]} '='  ${1#*=}"	;;
		
		# Integer comparisons
		*\<=*)		echo "(( \${platform_options[${1%%<*}]%%.*} '<=' ${1#*=} ))"		;;
		*\>=*)		echo "(( \${platform_options[${1%%>*}]%%.*} '>=' ${1#*=} ))"		;;
		*==*)		echo "(( \${platform_options[${1%%=*}]%%.*} '==' ${1#*=} ))"		;;
		*\<*)		echo "(( \${platform_options[${1%%<*}]%%.*} '<'  ${1#*<} ))"		;;
		*\>*)		echo "(( \${platform_options[${1%%>*}]%%.*} '>'  ${1#*>} ))"		;;
		*=*)		echo "(( \${platform_options[${1%%=*}]%%.*} '='  ${1#*=} ))"		;;
		
		# Presence test
		*)			echo "\${platform_options[$1]:-false}"							;;
	esac
}

#
# Parse a dependency list and provide a list of existing packages to load
#
# The input is a set of package dependencies, with an optional version tag (+ optional arch) or constraint:
#
#		package-name_package-version_package-arch
#		package-name_package-version
#		package-name
#		package-name (version-constraints)
#
# Where version-constraint can be one or more constraints:
#		< version
#		<= version
#		== version (can be abridged as '= version' or just 'version')
#		!= version
#		>= version
#		> version
#
# When several constraints are given, they may be separated by
#		||
#		&& (optional)
# Note that && has higher precedence than || and parentheses are not supported
#
parse_depend() {
	local required="$1"
	shift
	local dependencies="$*"
	dependencies="${dependencies//(/ ( }"
	dependencies="${dependencies//)/ ) }"
	dependencies="${dependencies//</ < }"
	dependencies="${dependencies//>/ > }"
	dependencies="${dependencies//=/ = }"
	dependencies="${dependencies//!/ ! }"
	dependencies="${dependencies//&/ & }"
	dependencies="${dependencies//|/ | }"
	dependencies="${dependencies//< =/<=}"
	dependencies="${dependencies//> =/>=}"
	dependencies="${dependencies//= =/=}"
	dependencies="${dependencies//! =/!=}"
	dependencies="${dependencies//& &/&&}"
	dependencies="${dependencies//| |/||}"
	declare -a tokens=($dependencies)
	pkg_dbg "dependencies='$dependencies'"
	local state=init
	local token next_token i j pkg pkg_spec='' 
	local versions accumulated_versions filtered version
	local i=0
	for ((i=0; i<=${#tokens[*]}; i++)); do
		if [[ $i -lt ${#tokens[*]} ]]; then
			token="${tokens[i]}"
			j=$((i+1))
			if [[ $j -lt ${#tokens[*]} ]]; then
				next_token="${tokens[j]}"
			else
				next_token="EOF" # End of dependencies sentinel
			fi
		else
			token="EOF"
			next_token="EOF"
		fi
		pkg_dbg "versions='$versions'"
		pkg_dbg "accumulated_versions='$accumulated_versions'"
		pkg_dbg "filtered='$filtered'"
		pkg_dbg "token='$token'"
		pkg_dbg "next_token='$next_token'"
		pkg_dbg "state='$state'"
		pkg_dbg "pkg_spec='$pkg_spec'"
		case $state in
			init | parsed)
				if [ $state = parsed ]; then
					if provide_version $required $pkg_spec $versions $filtered; then
						true
					else
						return 1
					fi
					if [ "$token" = "EOF" ]; then
						break
					fi
				fi

				pkg="$token"
				pkg_spec="$pkg"
				versions=
				for version in $(search_all_versions $pkg); do
					versions="$versions $version"
				done
				versions="$versions "
				case "$token" in
					[a-zA-Z0-9]*)
						case "$next_token" in
							[a-zA-Z0-9]*)
								state=parsed
								;;
							*)
								state=constrained
								;;
						esac
						;;
					*)	echo "Syntax error in '$dependencies' at '$pkg': not a package name" >&2
						return 1
						;;
				esac
				;;
			constrained)
				op="$token"
				case "$op" in
					\()	
						pkg_spec="$pkg_spec $op"
						((level++))
						# Here to manage parentheses I should manage to_filter, filtered
						# and accumulated_versions in a stack
						to_filter="$versions"
						accumulated_versions=
						filtered=
						state=operator
						;;
					*)
						echo "Syntax error in '$pkg_spec' dependency: invalid token '$op'" >&2
						return 1
						;;
				esac
				;;
			operator)
				cmp="$token"
				case "$cmp" in
					\))	
						pkg_spec="$pkg_spec $op"
						if [ $level -eq 0 ]; then
							echo "Syntax error in '$pkg_spec' dependency: too much closing parentheses" >&2
							return 1
						fi
						# Here to manage parentheses I should manage to_filter, filtered
						# and accumulated_versions in a stack
						((level--))
						if [ $level -eq 0 ]; then
							versions="$accumulated_versions $filtered"
							state=parsed
						fi
						;;
					'<' | '<=' | '==' | '=' | '!=' | '!' | '>=' | '>')
						state=operand
						;;
					'||')
						accumulated_versions="$accumulated_versions $filtered"
						to_filter="$versions"
						filtered=
						;;
					'&&')
						;;
					*)
						cmp='='
						state=operand
						((i--))
#						echo "Syntax error in '$pkg_spec' dependency: operator needed before '$cmp'" >&2
#						return 1
						;;
				esac
				;;
			operand)
				val="$token"
				filtered=
				for version in $to_filter; do
					vers=${version#*_}
					if version_expr "${vers%%_*}" $cmp "$val"; then
						filtered="$filtered $version"
					fi
				done
				to_filter="$filtered"
				;;
		esac
	done
}

#
# Analyze options provided in a filename to determine if it must be filtered out
# Will return true if the file must be filtered-out
# Will return false if the file must be used 
#
# If a second parameter is passed and it names an existing function, this function is
# called before analyzing each option; if it return true, the option is skipped.
#
# If a third parameter is passed and it names an existing function, this function is
# called if an option is not recognized as a filtering option; the option is then ignored
#
filter_on_options() {
	local fileName="$1"
	local pre_analyze_fn="$2"
	local pre_analyze=false
	local post_analyze_fn="$2"
	local post_analyze=false
	local optionsString="${fileName#*@}"
	local -a options=()
	local condition=""

	if [ -n "$pre_analyze_fn" ] && fn_exists "$pre_analyze_fn"; then
		pre_analyze=true
	fi
	if [ -n "$post_analyze_fn" ] && fn_exists "$post_analyze_fn"; then
		post_analyze=true
	fi

	if [ "$optionsString" != "$fileName" ]; then
		options=( ${optionsString/;/ } )
		local option comparison sep
		for option in "${options[@]}"; do
			if [ -z "$option" ]; then continue; fi
			if $pre_analyze && $pre_analyze_fn $option; then continue; fi
			case $option in
				when=* | and=*)
					comparison=$(get_comparison ${option#*=})
					if [ -n "$condition" -a -n "$comparison" ]; then sep=" && "; else sep=""; fi
					condition="$condition$sep$comparison"
					;;
				or=*)
					comparison=$(get_comparison ${option#*=})
					if [ -n "$condition" -a -n "$comparison" ]; then sep=" || "; else sep=""; fi
					condition="$condition$sep$comparison"
					;;
				*=* | *\<* | *\>*)
					comparison=$(get_comparison ${option})
					if [ -n "$condition" -a -n "$comparison" ]; then sep=" || "; else sep=""; fi
					condition="$condition$sep$comparison"
					;;
				*)
					if $post_analyze; then $post_analyze_fn "$option"; fi
					;;
			esac
		done
	fi

	# Filter with respect to optional conditions
	if [ -n "$condition" ]; then
		if eval $condition; then
			log "Condition verified: '$condition'"
		else
			log "Condition failed: '$condition'"
			return 0
		fi
	fi
	return 1
}
